/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.gitrepository.controler;

import org.springframework.data.domain.Page;
import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.RequestMapping;
import org.springframework.web.bind.annotation.RequestMethod;
import org.springframework.web.bind.annotation.RequestParam;
import ru.ulstu.extractor.core.OffsetablePageRequest;
import ru.ulstu.extractor.gitrepository.model.FilterForm;
import ru.ulstu.extractor.gitrepository.service.FilteringService;
import ru.ulstu.extractor.model.Commit;
import springfox.documentation.annotations.ApiIgnore;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

import static ru.ulstu.extractor.core.Route.FILTER_COMMITS;

@Controller
@ApiIgnore
public class GitFilteringController {
    private final static int DEFAULT_PAGE_SIZE = 20;
    private final FilteringService filteringService;

    public GitFilteringController(FilteringService filteringService) {
        this.filteringService = filteringService;
    }

    @RequestMapping(value = FILTER_COMMITS, method = RequestMethod.GET)
    public String listCommits(
            Model model,
            @RequestParam Optional<Integer> page,
            @RequestParam Optional<Integer> size,
            @RequestParam Optional<String> repositoryUrl,
            @RequestParam Optional<String> branchName,
            @RequestParam Optional<String> author,
            @RequestParam Optional<String> filter,
            @RequestParam Optional<Boolean> entity) {
        int currentPage = page.orElse(1);
        int pageSize = size.orElse(DEFAULT_PAGE_SIZE);

        String notEmptyRepositoryUrl = repositoryUrl.orElseThrow(() -> new RuntimeException("Url repository not present"));
        String notEmptyBranchName = branchName.orElseThrow(() -> new RuntimeException("Branch name not present"));

        Page<Commit> commitsPage = filteringService.getCommits(notEmptyRepositoryUrl,
                notEmptyBranchName,
                author.orElse(null),
                filter.orElse(null),
                entity.orElse(null),
                new OffsetablePageRequest(currentPage - 1, pageSize));
        int totalPages = commitsPage.getTotalPages();
        if (totalPages > 0) {
            List<Integer> pageNumbers = IntStream.rangeClosed(1, totalPages)
                    .boxed()
                    .collect(Collectors.toList());
            model.addAttribute("pageNumbers", pageNumbers);
        }
        FilterForm filterForm = new FilterForm();
        filterForm.setCommitsPage(commitsPage);
        filterForm.setBranchName(notEmptyBranchName);
        filterForm.setRepositoryUrl(notEmptyRepositoryUrl);
        filterForm.setAuthor(author.orElse(null));
        filterForm.setFilter(filter.orElse(null));
        filterForm.setEntity(entity.orElse(null));
        model.addAttribute("filterForm", filterForm);
        model.addAttribute("authors", filteringService.getRepositoryAuthors(
                notEmptyRepositoryUrl,
                notEmptyBranchName
        ));

        model.addAttribute("entityPresent", filteringService.getEntityPresent());
        return FILTER_COMMITS;
    }
}
