/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.gitrepository.controler;

import org.springframework.stereotype.Controller;
import org.springframework.ui.Model;
import org.springframework.web.bind.annotation.GetMapping;
import org.springframework.web.bind.annotation.RequestParam;
import ru.ulstu.extractor.branch.service.BranchService;
import ru.ulstu.extractor.commit.repository.CommitRepository;
import ru.ulstu.extractor.gitrepository.model.FilterForm;
import ru.ulstu.extractor.gitrepository.service.FilteringService;
import springfox.documentation.annotations.ApiIgnore;

import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static ru.ulstu.extractor.core.Route.STATISTIC;

@Controller
@ApiIgnore
public class StatisticController {
    private final CommitRepository commitRepository;
    private final FilteringService filteringService;
    private final BranchService branchService;

    public StatisticController(CommitRepository commitRepository,
                               FilteringService filteringService,
                               BranchService branchService) {
        this.commitRepository = commitRepository;
        this.filteringService = filteringService;
        this.branchService = branchService;
    }

    @GetMapping(STATISTIC)
    public String getStatistic(Model model,
                               @RequestParam Optional<Integer> branchId,
                               @RequestParam Optional<String> author,
                               @RequestParam Optional<String> filter,
                               @RequestParam Optional<Boolean> entity) {
        List<Object[]> urlCommits = getUrlCommits();
        List<Object[]> timeCommits = getTimeCommits(branchId, author);
        model.addAttribute("commitAuthorData", authorCommits(branchId, author));
        model.addAttribute("commitUrlData", urlCommits);
        model.addAttribute("commitTimeData", timeCommits);
        List<String> dates = timeCommits.stream()
                .map(tc -> tc[0].toString())
                .collect(Collectors.toList());
        model.addAttribute("dates", dates);

        List<String> urls = urlCommits.stream()
                .map(tc -> tc[0].toString().substring(tc[0].toString().lastIndexOf("/") + 1))
                .collect(Collectors.toList());
        model.addAttribute("urls", urls);
        model.addAttribute("commitEntityData", getEntityCommits(branchId, author));
        model.addAttribute("commitTimeEntityData", allTimeEntityCommits(branchId, author));
        FilterForm filterForm = new FilterForm();
        filterForm.setBranchId(branchId.orElse(null));
        filterForm.setAuthor(author.orElse(null));
        filterForm.setFilter(filter.orElse(null));
        filterForm.setEntity(entity.orElse(null));
        model.addAttribute("filterForm", filterForm);
        model.addAttribute("entityPresent", filteringService.getEntityPresent());
        model.addAttribute("branches", branchService.findAll());
        model.addAttribute("authors", filteringService.getRepositoryAuthors(branchId.orElse(null)));
        return STATISTIC;
    }

    public List<Object[]> authorCommits(Optional<Integer> branchId, Optional<String> author) {
        return commitRepository.getCommitAuthorStatistic(
                branchId.orElse(null),
                author.orElse(null))
                .stream()
                .map(stat -> new Object[]{stat.getAuthor(), stat.getCountCommit()})
                .collect(Collectors.toList());
    }

    public List<Object[]> getUrlCommits() {
        return commitRepository.getCommitUrlStatistic().stream()
                .map(stat -> new Object[]{stat.getUrl(), stat.getCountCommit()})
                .collect(Collectors.toList());
    }

    public List<Object[]> getTimeCommits(Optional<Integer> branchId, Optional<String> author) {
        return commitRepository.getCommitTimeStatistic(branchId.orElse(null), author.orElse(null)).stream()
                .map(stat -> new Object[]{stat.getDate(), stat.getCountCommit()})
                .collect(Collectors.toList());
    }

    public List<Object[]> getEntityCommits(Optional<Integer> branchId, Optional<String> author) {
        List<Object[]> entityCommits = commitRepository.getCommitEntityStatistic(branchId.orElse(null), author.orElse(null)).stream()
                .map(stat -> new Object[]{stat.getEntity(), stat.getCountCommit()})
                .collect(Collectors.toList());
        for (Object[] entityCommit : entityCommits) {
            entityCommit[0] = Boolean.TRUE.equals(entityCommit[0])
                    ? "Есть сущности"
                    : (entityCommit[0] == null
                    ? "Нет данных"
                    : "Нет сущностей");
        }
        return entityCommits;
    }

    public List<Object[]> getTimeEntityCommits(Optional<Integer> branchId, Optional<String> author) {
        return commitRepository.getCommitTimeEntityStatistic(branchId.orElse(null), author.orElse(null)).stream()
                .map(stat -> new Object[]{stat.getDate(), stat.getCountCommit()})
                .collect(Collectors.toList());
    }

    public List<Object[]> allTimeEntityCommits(Optional<Integer> branchId, Optional<String> author) {
        List<Object[]> timeEntityCommits = getTimeEntityCommits(branchId, author);
        List<Object[]> allTimeEntityCommits = commitRepository.getCommitTimeStatistic(branchId.orElse(null), author.orElse(null)).stream()
                .map(stat -> new Object[]{stat.getDate(), stat.getCountCommit()})
                .collect(Collectors.toList());
        int j = 0;
        for (int i = 0; i < allTimeEntityCommits.size(); i++) {
            if ((j < timeEntityCommits.size()) && (allTimeEntityCommits.get(i)[0].equals(timeEntityCommits.get(j)[0]))) {
                allTimeEntityCommits.get(i)[1] = timeEntityCommits.get(j)[1];
                j++;
            } else {
                if (i > 0) {
                    allTimeEntityCommits.get(i)[1] = allTimeEntityCommits.get(i - 1)[1];
                }
            }
        }
        return allTimeEntityCommits;
    }
}
