package ru.ulstu.extractor.gitrepository.service;

import org.apache.commons.io.FileUtils;
import org.eclipse.jgit.api.CreateBranchCommand;
import org.eclipse.jgit.api.Git;
import org.eclipse.jgit.api.ListBranchCommand;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.eclipse.jgit.diff.DiffFormatter;
import org.eclipse.jgit.internal.storage.file.FileRepository;
import org.eclipse.jgit.lib.ObjectId;
import org.eclipse.jgit.lib.ObjectLoader;
import org.eclipse.jgit.lib.ObjectReader;
import org.eclipse.jgit.revwalk.RevCommit;
import org.eclipse.jgit.treewalk.TreeWalk;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.domain.Page;
import org.springframework.data.domain.Pageable;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.author.model.Author;
import ru.ulstu.extractor.branch.model.Branch;
import ru.ulstu.extractor.gitrepository.model.FileChange;
import ru.ulstu.extractor.gitrepository.model.GitRepository;
import ru.ulstu.extractor.gitrepository.model.LineChange;
import ru.ulstu.extractor.gitrepository.repository.GitRepositoryRepository;
import ru.ulstu.extractor.heuristic.model.BusinessLogicUnit;
import ru.ulstu.extractor.heuristic.model.EntityUnit;
import ru.ulstu.extractor.heuristic.model.ResourceUnit;
import ru.ulstu.extractor.heuristic.service.StructuralUnitService;
import ru.ulstu.extractor.model.Commit;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URISyntaxException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.Optional;
import java.util.concurrent.CopyOnWriteArrayList;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;
import java.util.stream.Collectors;

import static org.apache.logging.log4j.util.Strings.isBlank;

@Service
public class GitRepositoryService {
    private final static Logger LOG = LoggerFactory.getLogger(GitRepositoryService.class);
    private static final String BRANCH_PREFIX = "refs/remotes/origin/";
    @Value("${extractor.custom-projects-dir}")
    private String customProjectsDir;
    private final ExecutorService executorService = Executors.newFixedThreadPool(8);
    private final ExecutorService executorServiceCommits = Executors.newFixedThreadPool(8);
    private final StructuralUnitService structuralUnitService;
    private final GitRepositoryRepository gitRepositoryRepository;

    public GitRepositoryService(StructuralUnitService structuralUnitService,
                                GitRepositoryRepository gitRepositoryRepository) {
        this.structuralUnitService = structuralUnitService;
        this.gitRepositoryRepository = gitRepositoryRepository;
    }

    public List<Branch> getRemoteBranches(String url) throws GitAPIException, IOException {
        LOG.debug("Get remote branches of {}. Clone", url);
        cloneOrUpdateRepo(url);
        LOG.debug("Get remote branches of {}. Get branches", url);
        org.eclipse.jgit.lib.Repository localRepo = new FileRepository(getProjectGitDirectory(url));
        Git git = new Git(localRepo);
        List<Branch> branches = git.branchList().setListMode(ListBranchCommand.ListMode.REMOTE)
                .call()
                .stream()
                .map(r -> new Branch(r.getName().replace(BRANCH_PREFIX, "")))
                .collect(Collectors.toList());
        git.close();
        localRepo.close();
        return branches;
    }

    public List<Branch> getLocalBranches(String url) throws GitAPIException, IOException {
        LOG.debug("Get local branches of {}. Clone", url);
        cloneOrUpdateRepo(url);
        LOG.debug("Get local branches of {}. Get branches", url);
        org.eclipse.jgit.lib.Repository localRepo = new FileRepository(getProjectGitDirectory(url));
        Git git = new Git(localRepo);
        List<Branch> branches = git.branchList()
                .call()
                .stream()
                .map(r -> new Branch(r.getName().replace(BRANCH_PREFIX, "")))
                .collect(Collectors.toList());
        git.close();
        localRepo.close();
        return branches;
    }

    public List<Commit> getCommits(String repositoryUrl, String branchName, int commitsFrom, int commitsTo, boolean needUpdate) throws GitAPIException, IOException {
        LOG.debug("Get commits of {}. Branch {}", repositoryUrl, branchName);
        if (needUpdate) {
            cloneOrUpdateRepo(repositoryUrl, branchName);
        }
        org.eclipse.jgit.lib.Repository localRepo = new FileRepository(getProjectGitDirectory(repositoryUrl));
        Git git = new Git(localRepo);

        List<RevCommit> commits = new ArrayList<>();
        git.log().setSkip(commitsFrom).setMaxCount(commitsTo - commitsFrom).call().forEach(commits::add);

        List<Commit> list = new CopyOnWriteArrayList<>();
        RevCommit prevCommit = null;
        LOG.debug("Start analyse {} commits", commits.size());
        LOG.debug(" {} to {} commits", commitsFrom, commitsTo);
        List<Future<Commit>> futureCommits = new ArrayList<>();
        for (RevCommit revCommit : commits) {
            final RevCommit forPrevCommit = prevCommit;
            futureCommits.add(executorServiceCommits.submit(() -> {
                Commit commit = new Commit(
                        revCommit.getFullMessage(),
                        new Author(revCommit.getAuthorIdent().getName()),
                        Date.from(Instant.ofEpochSecond(revCommit.getCommitTime())),
                        revCommit.getName());
                if (forPrevCommit != null) {
                    commit.setFileChanges(findDiffBetweenTwoRevisions(revCommit, forPrevCommit, localRepo));
                }
                list.add(commit);
                return commit;
            }));
            prevCommit = revCommit;
        }
        futureCommits.forEach(fc -> {
            try {
                fc.get();
            } catch (Exception e) {
                LOG.debug(e.getMessage());
            }
        });
        git.close();
        localRepo.close();
        return list;
    }

    private void checkoutBranch(String repositoryUrl, Git git, org.eclipse.jgit.lib.Repository localRepo, String branchName) throws GitAPIException, IOException {
        LOG.debug("Checkout branch {} {}", repositoryUrl, branchName);
        git.pull().call();
        if (!localRepo.getBranch().equals(branchName)) {
            if (getLocalBranches(repositoryUrl).stream().anyMatch(localBranch -> localBranch.getName().contains(branchName))) {
                git.checkout()
                        .setName(branchName)
                        .call();
            } else {
                git.checkout()
                        .setCreateBranch(true)
                        .setName(branchName)
                        .setUpstreamMode(CreateBranchCommand.SetupUpstreamMode.TRACK)
                        .setStartPoint("origin/" + branchName)
                        .call();
            }
        }
        git.pull().call();
    }

    public File getProjectDirectoryFile(String url) {
        validateUrl(url);
        return Path.of(getProjectDirectory(url))
                .toFile();
    }

    public List<EntityUnit> getEntities(String repositoryUrl, String branchName) throws IOException, GitAPIException {
        cloneOrUpdateRepo(repositoryUrl, branchName);
        return structuralUnitService.getEntities(getProjectDirectoryFile(repositoryUrl));
    }

    public List<BusinessLogicUnit> getBusinessLogic(String repositoryUrl, String branchName) throws IOException, GitAPIException {
        cloneOrUpdateRepo(repositoryUrl, branchName);
        return structuralUnitService.getBusinessLogic(getProjectDirectoryFile(repositoryUrl));
    }

    public List<ResourceUnit> getResource(String repositoryUrl, String branchName) throws IOException, GitAPIException {
        cloneOrUpdateRepo(repositoryUrl, branchName);
        return structuralUnitService.getResource(getProjectDirectoryFile(repositoryUrl));
    }


    public void remove(String repositoryUrl) throws IOException {
        FileUtils.deleteDirectory(getProjectDirectoryFile(repositoryUrl));
    }

    private void cloneOrUpdateRepo(String repositoryUrl, String branchName) throws GitAPIException, IOException {
        Git git;
        org.eclipse.jgit.lib.Repository localRepo;
        validateOrClearRepositoryDirectory(getProjectDirectoryFile(repositoryUrl));
        if (projectDirExists(getProjectDirectoryFile(repositoryUrl))) {
            localRepo = new FileRepository(getProjectGitDirectory(repositoryUrl));
            git = new Git(localRepo);
            git.pull().call();
            localRepo.close();
        } else {
            git = Git.cloneRepository()
                    .setURI(repositoryUrl)
                    .setDirectory(getProjectDirectoryFile(repositoryUrl))
                    .call();
            localRepo = new FileRepository(getProjectGitDirectory(repositoryUrl));
        }
        if (branchName != null && !branchName.isEmpty()) {
            checkoutBranch(repositoryUrl, git, localRepo, branchName);
        }
        git.close();
    }

    private void validateOrClearRepositoryDirectory(File projectDirectory) {
        if (projectDirectory.exists()) {
            try {
                Git.open(projectDirectory).status().call();
            } catch (Exception e) {
                deleteDirectory(projectDirectory);
            }
        }
    }

    private boolean deleteDirectory(File directoryToBeDeleted) {
        File[] allContents = directoryToBeDeleted.listFiles();
        if (allContents != null) {
            for (File file : allContents) {
                deleteDirectory(file);
            }
        }
        return directoryToBeDeleted.delete();
    }

    private void cloneOrUpdateRepo(String url) throws GitAPIException, IOException {
        cloneOrUpdateRepo(url, null);
    }

    private String getProjectDirectory(String url) {
        return (isBlank(customProjectsDir)
                ? System.getProperty("java.io.tmpdir")
                : customProjectsDir) + url.substring(url.lastIndexOf('/'));
    }

    private String getProjectGitDirectory(String url) {
        return getProjectDirectory(url) + "/.git";
    }

    private void validateUrl(String url) {
        if (url == null || url.isEmpty()) {
            throw new RuntimeException("Repository url must not empty");
        }
        try {
            new URL(url).toURI();
        } catch (MalformedURLException | URISyntaxException e) {
            throw new RuntimeException("Repository url not valid");
        }
    }

    private boolean projectDirExists(File file) {
        return file.exists();
    }

    private List<FileChange> findDiffBetweenTwoRevisions(RevCommit laterCommit, RevCommit earlierCommit, org.eclipse.jgit.lib.Repository localRepo) {
        if (laterCommit == null || earlierCommit == null) {
            return null;
        }
        String output;
        try (ByteArrayOutputStream out = new ByteArrayOutputStream()) {
            DiffFormatter diffFormatter = new DiffFormatter(out);
            diffFormatter.setRepository(localRepo);
            diffFormatter.format(earlierCommit.getId(), laterCommit.getId());
            output = out.toString();
        } catch (IOException e) {
            throw new RuntimeException("Error occurred during diff computation. Message: " + e.getMessage());
        }
        return parseOutputDiff(output, localRepo, laterCommit);
    }

    private List<FileChange> parseOutputDiff(String output, org.eclipse.jgit.lib.Repository repository, RevCommit commit) {
        List<FileChange> changes = new ArrayList<>();
        String[] strings = output.split("\n");
        FileChange fileChange = new FileChange();
        int stringsLength = strings.length - 1;
        for (int i = 0; i < strings.length; i++) {
            while (i < stringsLength) {
                Optional<String> maybeFileName = getFileName(strings[i]);
                if (maybeFileName.isPresent()) {
                    fileChange = new FileChange();
                    fileChange.setFile(maybeFileName.get());
                    Future<Boolean> futureEntity = executorService.submit(() -> structuralUnitService.containsEntity(getContent(repository, commit, maybeFileName.get())));
                    Future<Boolean> futureBL = executorService.submit(() -> structuralUnitService.containsBusinessLogic(getContent(repository, commit, maybeFileName.get())));
                    try {
                        fileChange.setContainsBusinessLogic(futureBL.get());
                        fileChange.setContainsEntity(futureEntity.get());
                    } catch (Exception ex) {
                        LOG.warn(ex.getMessage());
                    }
                    /// вытащить другие изменения из коммита
                    changes.add(fileChange);
                }
                LineChange lineChange = new LineChange();
                if (strings[i].startsWith("-")) {
                    while ((i < stringsLength) && strings[i].startsWith("-")) {
                        if (strings[i].startsWith("---")) {
                            i++;
                        } else {
                            if (lineChange.getLineFrom() == null) {
                                lineChange.setLineFrom(strings[i]);
                                i++;
                            } else {
                                lineChange.setLineFrom(lineChange.getLineFrom() + "\n" + strings[i]);
                                i++;
                            }
                        }
                    }
                    lineChange.setRemoved(true);
                }
                if (strings[i].startsWith("+")) {
                    while ((i < stringsLength) && strings[i].startsWith("+")) {
                        if (strings[i].startsWith("+++")) {
                            i++;
                        } else {
                            if (lineChange.getLineTo() == null) {
                                lineChange.setLineTo(strings[i]);
                                i++;
                            } else {
                                lineChange.setLineTo(lineChange.getLineTo() + "\n" + strings[i]);
                                i++;
                            }
                        }
                    }
                    lineChange.setAdded(true);
                }
                if ((lineChange.getLineTo() != null) || (lineChange.getLineFrom() != null)
                        || ((lineChange.getLineTo() != null) && (lineChange.getLineFrom() != null))) {
                    fileChange.getLineChanges().add(lineChange);
                } else {
                    i++;
                }
            }
        }
        return changes;
    }

    private String getContent(org.eclipse.jgit.lib.Repository repository, RevCommit commit, String path) {
        try (TreeWalk treeWalk = TreeWalk.forPath(repository, path, commit.getTree())) {
            if (treeWalk != null) {
                ObjectId blobId = treeWalk.getObjectId(0);
                try (ObjectReader objectReader = repository.newObjectReader()) {
                    ObjectLoader objectLoader = objectReader.open(blobId);
                    byte[] bytes = objectLoader.getBytes();
                    return new String(bytes, StandardCharsets.UTF_8);
                }
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return "";
    }

    private Optional<String> getFileName(String commitString) {
        String startString = "diff --git a/";
        if (commitString.startsWith(startString)) {
            String name = commitString.substring(commitString.indexOf(startString) + startString.length(), commitString.indexOf(" b/"));
            return Optional.of(name);
        }
        return Optional.empty();
    }

    public Page<GitRepository> findAll(Pageable pageable) {
        return gitRepositoryRepository.findAll(pageable);
    }

    public GitRepository findById(Integer id) {
        return gitRepositoryRepository.getOne(id);
    }
}
