/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.gitrepository.service;

import com.sun.istack.NotNull;
import org.eclipse.jgit.api.errors.GitAPIException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.branch.model.Branch;
import ru.ulstu.extractor.branch.service.BranchService;
import ru.ulstu.extractor.gitrepository.model.GitRepository;
import ru.ulstu.extractor.gitrepository.repository.GitRepositoryRepository;
import ru.ulstu.extractor.model.Commit;
import ru.ulstu.extractor.ts.creator.AbstractTimeSeriesCreator;

import java.io.IOException;
import java.util.Collections;
import java.util.List;

@Service
public class IndexService {
    private final static Logger LOG = LoggerFactory.getLogger(IndexService.class);
    private final static int COMMITS_PAGE_SIZE = 10;
    private final GitRepositoryService gitRepositoryService;
    private final GitRepositoryRepository gitRepositoryRepository;
    private final BranchService branchService;
    private final List<AbstractTimeSeriesCreator> timeSeriesCreators;

    public IndexService(GitRepositoryService gitRepositoryService,
                        GitRepositoryRepository gitRepositoryRepository,
                        BranchService branchService,
                        List<AbstractTimeSeriesCreator> timeSeriesCreators) {
        this.gitRepositoryService = gitRepositoryService;
        this.gitRepositoryRepository = gitRepositoryRepository;
        this.branchService = branchService;
        this.timeSeriesCreators = timeSeriesCreators;
    }

    public void index(@NotNull String repositoryUrl, @NotNull String branchName) throws GitAPIException, IOException {
        GitRepository gitRepository = gitRepositoryRepository.findByUrl(repositoryUrl);
        if (gitRepository == null) {
            gitRepository = gitRepositoryRepository.save(new GitRepository(repositoryUrl));
        }
        Branch branch = branchService.findByRepositoryAndName(gitRepository, branchName);
        if (branch == null) {
            branch = new Branch(gitRepository, branchName);
        }
        branchService.save(branch, Collections.emptyList());
        int commitsFrom = 0;
        int commitsTo = COMMITS_PAGE_SIZE;
        List<Commit> commits = gitRepositoryService.getCommits(repositoryUrl, branchName, commitsFrom, commitsTo, true);
        while (!commits.isEmpty()) {
            LOG.debug("{} commits loaded.", commits.size());
            branchService.addCommits(branch, commits);
            LOG.debug("{} commits successfully saved. {} {}", commits.size(), repositoryUrl, branchName);
            commitsFrom += COMMITS_PAGE_SIZE;
            commitsTo += COMMITS_PAGE_SIZE;
            commits = gitRepositoryService.getCommits(repositoryUrl, branchName, commitsFrom, commitsTo, false);
        }
        Integer repositoryId = gitRepository.getId();
        final Branch branchForSave = branch;
        timeSeriesCreators.forEach(tsCreator -> tsCreator.addTimeSeries(repositoryId, branchForSave));
        LOG.debug("Complete indexing {} branch", branchName);
    }
}
