/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.heuristic.api;

import ru.ulstu.extractor.heuristic.component.BuildTool;
import ru.ulstu.extractor.heuristic.component.ProgrammingLanguage;
import ru.ulstu.extractor.heuristic.model.BusinessLogicUnit;
import ru.ulstu.extractor.heuristic.model.EntityUnit;
import ru.ulstu.extractor.heuristic.model.ResourceUnit;
import ru.ulstu.extractor.heuristic.service.DetectorService;
import ru.ulstu.extractor.heuristic.service.ProgrammingLanguageService;
import ru.ulstu.extractor.util.StringUtils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import static ru.ulstu.extractor.heuristic.service.DetectorService.LangDetectScrupulousness.HIGH;
import static ru.ulstu.extractor.heuristic.service.DetectorService.LangDetectScrupulousness.LOW;

public abstract class StructuralUnitIdentifier {
    public List<EntityUnit> getEntityClasses(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        String subDirectory = getSourceDirectory(rootProjectFiles);
        return getEntityClasses(projectPath, subDirectory, projectFiles);
    }

    public List<BusinessLogicUnit> getBusinessLogicClasses(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        String subDirectory = getSourceDirectory(rootProjectFiles);
        return getBusinessLogicClasses(projectPath, subDirectory, projectFiles);
    }

    public List<ResourceUnit> getResourceClasses(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        String subDirectory = getSourceDirectory(rootProjectFiles);
        return getResourceClasses(projectPath, subDirectory, projectFiles);
    }

    public abstract boolean canAppliedToCode(String sourceCode);

    public abstract boolean canAppliedToFile(File projectFile);

    public abstract boolean canAppliedToProject(String projectPath, List<File> projectFiles, List<File> rootProjectFiles);

    public abstract boolean isEntityClass(File file);

    public abstract boolean isEntityClass(String sourceCode);

    public abstract boolean isBusinessLogicClass(String sourceCode);

    protected abstract boolean isBusinessLogicClass(File file);

    public abstract List<BusinessLogicUnit> getBusinessLogicClasses(String projectPath, String subDirectory, List<File> projectFiles);

    public abstract boolean isMultiModuleProject();

    public abstract Optional<BuildTool> getBuildTool(List<File> rootDirectoryFiles);

    protected abstract List<EntityUnit> getEntityClasses(String projectPath, String subDirectory, List<File> projectFiles);

    protected abstract String getSourceDirectory(List<File> rootProjectFiles);

    protected abstract DetectorService getDetectorService();

    protected abstract ProgrammingLanguageService getProgrammingLanguageService();

    protected abstract ProgrammingLanguage getProgrammingLanguage();

    protected Optional<ProgrammingLanguage> getMainProgrammingLanguage(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        String subDirectory = getSourceDirectory(rootProjectFiles);
        Map<ProgrammingLanguage, Integer> projectFileLanguageFrequency = new HashMap<>();
        projectFiles.stream()
                .filter(file -> StringUtils.fileInSubdirectory(file.getPath(), projectPath, subDirectory))
                .forEach(projectFile -> {
                    Optional<ProgrammingLanguage> detectedLanguage = getMainProgrammingLanguage(projectFile, LOW);
                    detectedLanguage.ifPresent(programmingLanguage -> projectFileLanguageFrequency.put(programmingLanguage, projectFileLanguageFrequency.getOrDefault(programmingLanguage, 0) + 1));
                });

        Optional<Map.Entry<ProgrammingLanguage, Integer>> mostFrequentLanguageEntry = projectFileLanguageFrequency
                .entrySet()
                .stream()
                .max(Map.Entry.comparingByValue());
        if (mostFrequentLanguageEntry.isEmpty()) {
            return Optional.empty();
        }

        return Optional.of(mostFrequentLanguageEntry.get().getKey());
    }

    protected Optional<ProgrammingLanguage> getMainProgrammingLanguage(File projectFile, DetectorService.LangDetectScrupulousness scrupulousness) {
        String fileContent = "";
        if (scrupulousness == HIGH) {
            try {
                fileContent = new String(Files.readAllBytes(projectFile.toPath()));
            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        return getProgrammingLanguageService().getByName(getDetectorService().getDetectedLanguage(
                projectFile.getName(),
                fileContent,
                scrupulousness));
    }

    protected abstract List<ResourceUnit> getResourceClasses(String projectPath, String subDirectory, List<File> projectFiles);

    public abstract boolean isResourceClass(String sourceCode);

    protected abstract boolean isResourceClass(File file);
}
