/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.heuristic.service;

import com.github.javaparser.JavaParser;
import com.github.javaparser.ParseResult;
import com.github.javaparser.ast.CompilationUnit;
import com.github.javaparser.ast.body.FieldDeclaration;
import com.github.javaparser.ast.body.MethodDeclaration;
import com.github.javaparser.ast.body.TypeDeclaration;
import org.springframework.stereotype.Service;
import ru.ulstu.extractor.heuristic.api.StructuralUnitIdentifier;
import ru.ulstu.extractor.heuristic.component.BuildTool;
import ru.ulstu.extractor.heuristic.component.JavaProgrammingLanguage;
import ru.ulstu.extractor.heuristic.component.ProgrammingLanguage;
import ru.ulstu.extractor.heuristic.model.BusinessLogicUnit;
import ru.ulstu.extractor.heuristic.model.EntityUnit;
import ru.ulstu.extractor.heuristic.model.ResourceUnit;
import ru.ulstu.extractor.util.StringUtils;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

import static ru.ulstu.extractor.heuristic.service.DetectorService.LangDetectScrupulousness.LOW;

@Service
public class JavaIdentifier extends StructuralUnitIdentifier {
    private static final String ENTITY_ANNOTATION = "@Entity";
    private static final String SERVICE_ANNOTATION = "@Service";
    private final DetectorService detectorService;
    private final BuildToolService buildToolService;
    private final ProgrammingLanguageService programmingLanguageService;

    public JavaIdentifier(DetectorService detectorService,
                          BuildToolService buildToolService,
                          ProgrammingLanguageService programmingLanguageService) {
        this.detectorService = detectorService;
        this.buildToolService = buildToolService;
        this.programmingLanguageService = programmingLanguageService;
    }

    @Override
    public boolean canAppliedToProject(String projectPath, List<File> projectFiles, List<File> rootProjectFiles) {
        return /*getBuildTool() instanceof GradleBuildTool
                &&*/ getMainProgrammingLanguage(projectPath, projectFiles, rootProjectFiles).orElse(null) instanceof JavaProgrammingLanguage;
    }

    public boolean canAppliedToFile(File projectFile) {
        return getMainProgrammingLanguage(projectFile, LOW).orElse(null) instanceof JavaProgrammingLanguage;
    }

    public boolean canAppliedToCode(String sourceCode) {
        return getMainProgrammingLanguage(sourceCode).orElse(null) instanceof JavaProgrammingLanguage;
    }

    @Override
    public List<EntityUnit> getEntityClasses(String projectPath, String subDirectory, List<File> projectFiles) {
        return projectFiles.stream()
                .filter(file -> StringUtils.fileInSubdirectory(file.getPath(), projectPath, subDirectory))
                .filter(this::isEntityClass)
                .map(file -> new EntityUnit(projectPath, file))
                .collect(Collectors.toList());
    }


    @Override
    public List<BusinessLogicUnit> getBusinessLogicClasses(String projectPath, String subDirectory, List<File> projectFiles) {
        return projectFiles.stream()
                .filter(file -> StringUtils.fileInSubdirectory(file.getPath(), projectPath, subDirectory))
                .filter(this::isBusinessLogicClass)
                .map(file -> new BusinessLogicUnit(projectPath, file))
                .collect(Collectors.toList());
    }

    @Override
    public boolean isMultiModuleProject() {
        return false;
    }

    @Override
    public Optional<BuildTool> getBuildTool(List<File> rootDirectoryFiles) {
        return buildToolService.getProjectBuildTool(rootDirectoryFiles);
    }

    protected Optional<ProgrammingLanguage> getMainProgrammingLanguage(String sourceCode) {
        return sourceCodeContainsClass(sourceCode)
                ? Optional.of(getProgrammingLanguage())
                : Optional.empty();
    }

    @Override
    protected DetectorService getDetectorService() {
        return detectorService;
    }

    @Override
    protected ProgrammingLanguageService getProgrammingLanguageService() {
        return programmingLanguageService;
    }

    @Override
    protected ProgrammingLanguage getProgrammingLanguage() {
        return new JavaProgrammingLanguage();
    }

    @Override
    protected List<ResourceUnit> getResourceClasses(String projectPath, String subDirectory, List<File> projectFiles) {
        return projectFiles.stream()
                .filter(file -> StringUtils.fileInSubdirectory(file.getPath(), projectPath, subDirectory))
                .filter(this::isResourceClass)
                .map(file -> new ResourceUnit(projectPath, file))
                .collect(Collectors.toList());
    }

    @Override
    public boolean isResourceClass(String sourceCode) {
        return false;
    }

    @Override
    public boolean isEntityClass(File file) {
        try {
            return file.getName().endsWith("java")
                    && (classContainsGetSet(file)
                    || classContainsAnnotation(file, ENTITY_ANNOTATION)
                    || packageContains(file, "model")
                    || packageContains(file, "entity"));
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }

    @Override
    public boolean isEntityClass(String sourceCode) {
        try {
            return classContainsAnnotation(sourceCode, ENTITY_ANNOTATION);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }

    @Override
    public boolean isBusinessLogicClass(String sourceCode) {
        try {
            return classContainsAnnotation(sourceCode, SERVICE_ANNOTATION);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }

    private boolean classContainsAnnotation(File file, String annotationDeclaration) throws IOException {
        return classContainsAnnotation(new String(Files.readAllBytes(file.toPath())), annotationDeclaration);
    }

    private boolean classContainsAnnotation(String sourceCode, String annotationDeclaration) {
        JavaParser parser = new JavaParser();
        ParseResult<CompilationUnit> parseResult = parser.parse(sourceCode);
        if (parseResult.getResult().isPresent() && parseResult.getResult().get().findCompilationUnit().isPresent()) {
            return parseResult.getResult().get().getTypes().stream()
                    .anyMatch(clazz -> clazz.getAnnotations().stream().anyMatch(annotation -> annotation.toString().startsWith(annotationDeclaration)));
        }
        return false;

    }

    private boolean sourceCodeContainsClass(String sourceCode) {
        JavaParser parser = new JavaParser();
        ParseResult<CompilationUnit> parseResult = parser.parse(sourceCode);
        if (parseResult.getResult().isPresent() && parseResult.getResult().get().findCompilationUnit().isPresent()) {
            return parseResult.getResult().get().getTypes().stream().findAny().isPresent();
        }
        return false;
    }

    @Override
    protected String getSourceDirectory(List<File> rootDirectoryFiles) {
        return buildToolService.getProjectBuildTool(rootDirectoryFiles)
                .map(BuildTool::getSourceDirectoryPath)
                .orElse("src");
    }

    @Override
    protected boolean isBusinessLogicClass(File file) {
        try {
            return file.getName().endsWith("java") && classContainsAnnotation(file, SERVICE_ANNOTATION);
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }

    private boolean classContainsGetSet(File file) throws IOException {
        return classContainsGetSet(new String(Files.readAllBytes(file.toPath())));
    }

    private boolean classContainsGetSet(String sourceCode) {
        JavaParser parser = new JavaParser();
        ParseResult<CompilationUnit> parseResult = parser.parse(sourceCode);
        if (parseResult.getResult().isPresent() && parseResult.getResult().get().findCompilationUnit().isPresent()) {
            for (TypeDeclaration<?> clazz : parseResult.getResult().get().getTypes()) {
                int countGetters = 0;
                int countSetters = 0;
                for (FieldDeclaration fieldDeclaration : clazz.getFields()) {
                    for (MethodDeclaration methodDeclaration : clazz.getMethods()) {
                        String fieldName = fieldDeclaration.getVariables().getFirst().map(v -> v.getName().toString()).orElse("").toUpperCase();
                        if (("GET" + fieldName).equals(methodDeclaration.getName().toString().toUpperCase())) {
                            countGetters++;
                        }
                        if (("SET" + fieldName).equals(methodDeclaration.getName().toString().toUpperCase())) {
                            countSetters++;
                        }
                    }
                }
                return clazz.getFields().size() == countGetters || clazz.getFields().size() == countSetters;
            }
        }
        return false;
    }

    private boolean packageContains(File file, String packageName) throws IOException {
        return packageContains(new String(Files.readAllBytes(file.toPath())), packageName);
    }

    private boolean packageContains(String sourceCode, String packageName) {
        JavaParser parser = new JavaParser();
        ParseResult<CompilationUnit> parseResult = parser.parse(sourceCode);
        if (parseResult.getResult().isPresent() && parseResult.getResult().get().findCompilationUnit().isPresent()) {
            return parseResult.getResult().isPresent() && parseResult.getResult().get().getPackageDeclaration().isPresent() && parseResult.getResult().get().getPackageDeclaration().get().getName().toString().contains(packageName);
        }
        return false;
    }

    protected boolean isResourceClass(File file) {
        try {
            return file.getName().endsWith("html") || file.getName().endsWith("css");
        } catch (Exception ex) {
            ex.printStackTrace();
        }
        return false;
    }
}
