/*
 * Copyright (C) 2021 Anton Romanov - All Rights Reserved
 * You may use, distribute and modify this code, please write to: romanov73@gmail.com.
 */

package ru.ulstu.extractor.heuristic.service;

import org.springframework.stereotype.Service;
import ru.ulstu.extractor.heuristic.api.StructuralUnitIdentifier;
import ru.ulstu.extractor.heuristic.model.BusinessLogicUnit;
import ru.ulstu.extractor.heuristic.model.EntityUnit;
import ru.ulstu.extractor.heuristic.model.ResourceUnit;

import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Optional;
import java.util.function.Predicate;

@Service
public class StructuralUnitService {
    private final DirectoryService directoryService;
    private final List<StructuralUnitIdentifier> structuralUnitIdentifiers;

    public StructuralUnitService(DirectoryService directoryService,
                                 List<StructuralUnitIdentifier> structuralUnitIdentifiers) {
        this.directoryService = directoryService;
        this.structuralUnitIdentifiers = structuralUnitIdentifiers;
    }

    public List<EntityUnit> getEntities(File rootPath) throws IOException {
        List<File> projectFiles = directoryService.getFilesRecursively(rootPath);
        List<File> rootProjectFiles = directoryService.getDirectoryFiles(rootPath.toPath());
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToProject(
                        rootPath.getPath(),
                        projectFiles,
                        rootProjectFiles))
                .orElseThrow(() -> new RuntimeException("Identifier not found"))
                .getEntityClasses(rootPath.getPath(), projectFiles, rootProjectFiles);
    }

    public List<BusinessLogicUnit> getBusinessLogic(File rootPath) throws IOException {
        List<File> projectFiles = directoryService.getFilesRecursively(rootPath);
        List<File> rootProjectFiles = directoryService.getDirectoryFiles(rootPath.toPath());
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToProject(
                        rootPath.getPath(),
                        projectFiles,
                        rootProjectFiles))
                .orElseThrow(() -> new RuntimeException("Identifier not found"))
                .getBusinessLogicClasses(rootPath.getPath(), projectFiles, rootProjectFiles);
    }

    public boolean containsEntity(File projectFile) {
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToFile(projectFile))
                .map(identifier -> identifier.isEntityClass(projectFile))
                .orElse(false);
    }

    public boolean containsEntity(String sourceCode) {
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToCode(sourceCode))
                .map(identifier -> identifier.isEntityClass(sourceCode))
                .orElse(false);
    }

    private Optional<StructuralUnitIdentifier> getStructuralUnitIdentifier(Predicate<StructuralUnitIdentifier> predicate) {
        return structuralUnitIdentifiers.stream()
                .filter(predicate)
                .findAny();
    }

    public boolean containsBusinessLogic(String sourceCode) {
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToCode(sourceCode))
                .map(identifier -> identifier.isBusinessLogicClass(sourceCode))
                .orElse(false);
    }

    public List<ResourceUnit> getResource(File rootPath) throws IOException {
        List<File> projectFiles = directoryService.getFilesRecursively(rootPath);
        List<File> rootProjectFiles = directoryService.getDirectoryFiles(rootPath.toPath());
        return getStructuralUnitIdentifier(
                structuralUnitIdentifier -> structuralUnitIdentifier.canAppliedToProject(
                        rootPath.getPath(),
                        projectFiles,
                        rootProjectFiles))
                .orElseThrow(() -> new RuntimeException("Identifier not found"))
                .getResourceClasses(rootPath.getPath(), projectFiles, rootProjectFiles);
    }
}
